/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "demodescription.h"
#include "gdamchannel.h"
#include <stdio.h>
#include <gtk/gtk.h>

extern DemoDescription demo_descriptions[];
extern int num_demo_descriptions;

static GdamChannel* channel;
static guint mixer_id;
static guint device_id;

static void run_demo_by_number(gint index)
{
	g_assert(0 <= index && index < num_demo_descriptions);
	demo_description_run(demo_descriptions + index, channel, mixer_id);
}

static void usage(const char* prog_name)
{
	fprintf(stderr, "usage: %s [gdam-options]\n\n"
	                "Gives a little gdam demo.\n\n"
			"Gdam-Options:\n"
			"%s",
			prog_name,
			gdam_channel_usage_message);
	exit(1);
}

int main(int argc, char** argv)
{
	int i;
	GtkWidget* window;
	GtkWidget* vbox;

	if (argc >= 2) {
		if (strcmp(argv[1], "--help") == 0
		 || strcmp(argv[1], "-help") == 0
		 || strcmp(argv[1], "-h") == 0)
		 	usage(argv[0]);
	}
	if (!gdam_channel_init(&argc, &argv,
	                       &channel)) {
		g_error("error creating connection to client");
		return 1;
	}
	g_warning("INCOMPLETE:  haven't yet set up device_id or mixer_id");
	g_return_val_if_fail(device_id != 0, 1);
	g_return_val_if_fail(mixer_id != 0, 1);
	gtk_init(&argc, &argv);

	if (argc != 1) {
		if (strcmp(argv[1], "--help") != 0
		 && strcmp(argv[1], "-help") != 0
		 && strcmp(argv[1], "-h") != 0)
			fprintf(stderr, "unknown option: %s\n\n", argv[1]);
		usage(argv[0]);
	}

	vbox = gtk_vbox_new(FALSE, 0);
	for (i = 0; i < num_demo_descriptions; i++) {
		DemoDescription* desc = demo_descriptions + i;
		GtkWidget* button;
		button = gtk_button_new_with_label(desc->description);
		gtk_signal_connect_object(GTK_OBJECT(button),
		                          "clicked",
				          GTK_SIGNAL_FUNC(run_demo_by_number),
					  (GtkObject*)i);
		gtk_box_pack_end_defaults((GtkBox*)vbox, button);
		gtk_widget_show(button);
	}
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_container_add(GTK_CONTAINER(window), GTK_WIDGET(vbox));
	gtk_widget_show(GTK_WIDGET(vbox));
	gtk_widget_show(window);
	gtk_main();
	return 0;
}
