/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamproducerqueueskin.h"
#include <gdam/component/gdamsongdatabase.h>
#include <gtk/gtkdnd.h>
#include <gtk/gtkclist.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkentry.h>
#include <glade/glade.h>
#include "../gdamskinutil.h"
#include "../gdamsongselectorskin.h"
#include "../gdampacker.h"
#include <string.h>
#include <stdlib.h>

/* TODO: add column fillers ala songsel, add column to mark current song */

static gboolean 
gdam_producer_queue_build_new_model(GdamSkin*	skin,
					GdamChannel*	channel,
					int 	 	num_options,
					char** 		options);
static void refresh_track_list(GdamProducerQueueSkin* skin,
			      GList*		      songinfo_list);
static void gdam_producer_queue_skin_class_init(GdamSkinClass* klass);
/*    ____ _                 ____  _          __  __ 
 *   / ___| | __ _ ___ ___  / ___|| |_ _   _ / _|/ _|
 *  | |   | |/ _` / __/ __| \___ \| __| | | | |_| |_ 
 *  | |___| | (_| \__ \__ \  ___) | |_| |_| |  _|  _|
 *   \____|_|\__,_|___/___/ |____/ \__|\__,_|_| |_|  
 */                                                 

static GtkObjectClass* parent_class = NULL;

static void gdam_producer_queue_skin_init(GdamProducerQueueSkin* skin)
{
}

GtkType gdam_producer_queue_skin_get_type()
{
	static GtkType producerqueueskin_type = 0;
	if (!producerqueueskin_type) {
		static const GtkTypeInfo producerqueueskin_info = {
			"GdamProducerQueueSkin",
			sizeof(GdamProducerQueueSkin),
			sizeof(GdamProducerQueueSkinClass),
			(GtkClassInitFunc) gdam_producer_queue_skin_class_init,
			(GtkObjectInitFunc) gdam_producer_queue_skin_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_PRODUCERMP3SKIN;
		parent_class = gtk_type_class(parent);
		producerqueueskin_type = gtk_type_unique(parent,
		                                     &producerqueueskin_info);

	}
	return producerqueueskin_type;
}

typedef struct _SkipInfo {
	GdamProducerQueue* queue;
	int num;
} SkipInfo;

/* A simple callback which traps the state of the most recent button_press callback */
static gboolean any_pressed_callback(GtkObject*         object,
                                     GdkEventButton*    event,
				     GdamProducerQueueSkin* skin)
{
	skin->event_state = event->state;
	return FALSE;
}


static void repeat_toggled(GdamProducerQueueSkin* skin, GtkWidget* widget)
{
	GtkToggleButton* toggle_button = GTK_TOGGLE_BUTTON(widget);
	GdamProducerQueue* queue = GDAM_PRODUCER_QUEUE(GDAM_SKIN(skin)->model);
	g_return_if_fail(toggle_button != NULL);
	g_return_if_fail(queue != NULL);
	queue->do_repeat = gtk_toggle_button_get_active(toggle_button);
	gdam_debug("queue repeat toggled.");
}

static void toggle_queue_visibility(GdamProducerQueueSkin* skin)
{
	if (!skin->queue_toplevel){
		gdam_debug("queue visibility couldn't be toggled.");
		return;
	}
	if (GTK_WIDGET_VISIBLE(skin->queue_toplevel))
		gtk_widget_hide(skin->queue_toplevel);
	else
		gtk_widget_show(skin->queue_toplevel);
	gdam_debug("queue visibility toggled.");
}

static void on_skip_clicked(GtkObject*	object,
			    SkipInfo*   skip_info)
{
	gdam_debug("SKIP CLICKED");
	gdam_producer_queue_print_queue(GDAM_PRODUCER_QUEUE(skip_info->queue));
	gdam_producer_queue_skip_songs(GDAM_PRODUCER_QUEUE(skip_info->queue), skip_info->num);
}

static gint sort_selection_info(gconstpointer a, gconstpointer b)
{
	if (GPOINTER_TO_INT(a) < GPOINTER_TO_INT(b)){
		return -1;
	}
	else if (GPOINTER_TO_INT(a) > GPOINTER_TO_INT(b)){
		return 1;
	}
	return 0;
}

static GList* get_sorted_selection_info(GdamProducerQueueSkin* skin)
{
	GList* selection;
	selection = g_list_copy(skin->track_list->selection);
	selection = g_list_sort(selection, sort_selection_info);
	return selection;
}

/* TODO: these 3 funcs should sort selection lists. */
static void on_delete_clicked(GtkObject* object,
			      GdamProducerQueueSkin* skin)
{
	GdamProducerQueue* queue;
	GtkCList* clist;
	GList* selection;
	GList* selected;
	queue = GDAM_PRODUCER_QUEUE(GDAM_SKIN(skin)->model);
	clist = skin->track_list;
	selection = get_sorted_selection_info(skin);
	selected = g_list_last(selection);

	gdam_debug("queue delete called.");

	while (selected)
	{
		int num;
		num = GPOINTER_TO_INT(selected->data);
		gdam_debug("queue delete: %d selected.", num);
		gdam_producer_queue_delete_range(queue, num+1, num+1);
		selected = selected->prev;
	}
	refresh_track_list(skin, queue->songinfo_list);
}

static void on_up_clicked(GtkObject* object,
			  GdamProducerQueueSkin* skin)
{
	GdamProducerQueue* queue;
	GtkCList* clist;
	GList* selection;
	GList* selected;
	queue = GDAM_PRODUCER_QUEUE(GDAM_SKIN(skin)->model);
	clist = skin->track_list;
	selection = get_sorted_selection_info(skin);
	selected = g_list_last(selection);

	gdam_debug("queue_up clicked. ");

	while (selected)
	{
		int num;
		num = GPOINTER_TO_INT(selected->data);
		gdam_producer_queue_move_range(queue, num+1, num+1, num);
		selected = selected->prev;
	}
}

static void on_refresh_clicked(GtkObject* 	      object,
			       GdamProducerQueueSkin* skin)
{
	GdamProducerQueue* queue;
	queue = GDAM_PRODUCER_QUEUE(GDAM_SKIN(skin)->model);
	refresh_track_list(skin, queue->songinfo_list);
}

static void on_select_clicked (GtkObject* 	      object,
			       GdamProducerQueueSkin* skin)
{
	GdamProducerQueue* queue;
	queue = GDAM_PRODUCER_QUEUE(GDAM_SKIN(skin)->model);
	if (skin->track_list->selection){
		int sel_idx;
		sel_idx = GPOINTER_TO_INT(skin->track_list->selection->data);
		gdam_producer_queue_skip_to(queue, sel_idx+1);
	}
}

static void on_down_clicked(GtkObject* object,
			  GdamProducerQueueSkin* skin)
{
	GdamProducerQueue* queue;
	GtkCList* clist;
	GList* selected;
	GList* selection;
	queue = GDAM_PRODUCER_QUEUE(GDAM_SKIN(skin)->model);
	clist = skin->track_list;
	selection = get_sorted_selection_info(skin);
	selected = selection;

	gdam_debug("queue_down clicked. ");

	while (selected)
	{
		int num;
		num = GPOINTER_TO_INT(selected->data);
		gdam_producer_queue_move_range(queue, num+1, num+1, num+2);
		selected = selected->next;
	}
}

/* Connects the clicked signal of the button to the handler, and also connects 
 * to the button_press_event signal, in order to trap the state of the event 
 * which produced the click. */
static void connect_button_with_mask(GdamProducerQueueSkin* skin, 
                                     GtkObject*         button,
				     GtkSignalFunc      clicked_callback,
				     gpointer           user_data)
{
	gtk_signal_connect(button,
	                   "clicked",
			   clicked_callback,
			   user_data);
	gtk_signal_connect(button,
	                   "button_press_event",
			   GTK_SIGNAL_FUNC(any_pressed_callback),
			   skin);
}


static void on_filename_received(
			    const char*   filename_str,
			    gpointer      data)
{
	GdamProducerQueueSkin* skin = GDAM_PRODUCER_QUEUE_SKIN(data);
	GdamProducerQueue* queue;
	char* nl;
	char* cur_file;
	char* filenames = g_strdup(filename_str);
	gboolean keep_going = TRUE;
	queue = GDAM_PRODUCER_QUEUE(GDAM_SKIN(skin)->model);
	cur_file = filenames;
	while (keep_going && *cur_file != '\0'){
		nl = strchr(cur_file, '\n');
		if (nl)
			*nl = '\0';
		else
			keep_going = FALSE;
		gdam_debug("filename %s dropped", cur_file);
		gdam_producer_queue_add_filename(queue, cur_file);
		cur_file = nl+1;
	}
	  /* Needed?  Not if args were used correctly... */
	refresh_track_list(skin, queue->songinfo_list);
	return;
}

/* Recreates the clist widget so that matches the given songinfo list. */
static void refresh_track_list(GdamProducerQueueSkin* skin,
			      GList*		      songinfo_list)
{
	GList* curlink = songinfo_list;
	GList* old_selection_data = NULL;
	GList* selected;
	GdamProducerQueue* queue;
	queue = GDAM_PRODUCER_QUEUE(GDAM_SKIN(skin)->model);

	gdam_debug("refreshing track list.");
	
	/* Save info about current selection */
	selected = skin->track_list->selection;
	while (selected)
	{
		int num;
		gpointer data;
		num = GPOINTER_TO_INT(selected->data);
		data = gtk_clist_get_row_data(skin->track_list, num);
		old_selection_data = g_list_append(old_selection_data, data);
		selected = selected->next;
	}

	/* Update clist to reflect contents of songinfo list */
	gtk_clist_freeze(skin->track_list);
	gtk_clist_clear(skin->track_list);
	while (curlink){
		GdamSongInfo* si = (GdamSongInfo*)curlink->data;
		char** strings;
		int i, new_idx;
		int num_columns = skin->track_list->columns;
		strings = (char**)alloca(sizeof(char*) * num_columns);
		for (i=0; i<num_columns; i++) {
			strings[i] = (*skin->column_fillers[i])(skin, si);
		}
		//strings[0] = g_strdup(gdam_song_info_get_filename(si));
		new_idx = gtk_clist_append(skin->track_list, strings);
		gtk_clist_set_row_data(skin->track_list, 
				       new_idx, 
				       curlink->data);
		curlink = curlink->next;
	}
	gtk_clist_thaw(skin->track_list);

	/* Restore selction state to degree possible. */
	selected = old_selection_data;
	while (selected)
	{
		int n;
		n = gtk_clist_find_row_from_data(skin->track_list, 
						 selected->data);
		if (n > 0)
			gtk_clist_select_row(skin->track_list, n, 0);
		selected = selected->next;
	}
}

static gboolean gdam_queue_filename_drag_received_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gdam_widget_string_dest_set(GTK_WIDGET(object),
				    on_filename_received,
				    "gdam-filename-text",
			            skin);
	return TRUE;
}


static gboolean do_clist_row_move(gpointer data)
{
	GdamProducerQueueSkin* skin = GDAM_PRODUCER_QUEUE_SKIN(data);
	GdamProducerQueue* queue;
	GtkCList* clist;
	GList* songinfo_list = NULL;
	int n;
	queue = GDAM_PRODUCER_QUEUE(GDAM_SKIN(skin)->model);
	clist = skin->track_list;

	gdam_debug("queue clist_row_move.");

	for (n=0; n<clist->rows; n++)
	{
		gpointer data;
		data = gtk_clist_get_row_data(clist, n);
		songinfo_list = g_list_append(songinfo_list, data);
	}

	gdam_producer_queue_set_queue_list(queue, songinfo_list);
	refresh_track_list(skin, songinfo_list);
	return FALSE;
}

static void on_clist_row_move(GtkObject*  	  object)
{
	g_idle_add(do_clist_row_move, object);
}

static void on_queue_list_set(GtkObject*	     object,
			      gpointer*		     list_pointer,
			      GdamProducerQueueSkin* skin)
{
	GList* songinfo_list = (GList*)list_pointer;
	gdam_debug("queue skin on_queue_list_set");
	refresh_track_list(skin, songinfo_list);
}

static gboolean is_one_of(const char* key, ...)
{
	va_list args;
	char* cmp;
	va_start(args, key);
	while ((cmp = va_arg(args, char*)) != NULL) {
		if (g_strcasecmp(key, cmp) == 0) {
			va_end(args);
			return TRUE;
		}
	}
	va_end(args);
	return FALSE;
}


static char* song_bpm_getter(GdamProducerQueueSkin* skin, GdamSongInfo* info)
{
	return g_strdup_printf("%3.3f", 
		gdam_beat_info_get_bpm(gdam_song_info_get_beat_info(info),0));
}

static char* song_queued_getter(GdamProducerQueueSkin* skin, GdamSongInfo* info)
{
	GdamProducerQueue* queue;
	queue = GDAM_PRODUCER_QUEUE(GDAM_SKIN(skin)->model);
	if (info == queue->current_song)
		return g_strdup_printf("=>"); 
	else
		return g_strdup_printf(".."); 
}

static char* song_length_getter(GdamProducerQueueSkin* skin, GdamSongInfo* info)
{
	int secs = (int)(gdam_song_info_get_length(info));
	return g_strdup_printf("%d:%02d", secs/60, secs%60);
}

static char* song_filename_getter(GdamProducerQueueSkin* skin, GdamSongInfo* info)
{
	return g_strdup(gdam_song_info_get_filename(info));
}

static char* unknown_getter(GdamProducerQueueSkin* skin, GdamSongInfo* info)
{
	return g_strdup("??");
}

static QListEntryGetter get_column_filler(const char* title)
{
	if (is_one_of(title, "filename",
	                     "song name", (char*) NULL))
		return song_filename_getter;

	if (is_one_of(title, "bpm", 
			     "tempo",
			     (char*) NULL))
		return song_bpm_getter;

	if (is_one_of(title, "length", 
			     (char*) NULL))
		return song_length_getter;

	if (is_one_of(title, "queued", 
			     (char*) NULL))
		return song_queued_getter;

	return unknown_getter;
}

static GtkJustification get_column_justification(const char* title)
{
	if (is_one_of(title, "filename",
	                     "song name", (char*) NULL))
		return GTK_JUSTIFY_LEFT;

	if (is_one_of(title, "bpm", 
			     "tempo",
			     (char*) NULL))
		return GTK_JUSTIFY_RIGHT;

	if (is_one_of(title, "length", 
			     (char*) NULL))
		return GTK_JUSTIFY_RIGHT;

	return GTK_JUSTIFY_LEFT;
}


/* column_names is ignored for now. */
static void connect_clist(GdamProducerQueueSkin* skin,
			  char** 		 column_names,
			  GtkCList*		 clist)
{
	GdamProducerQueue* queue;
	int i, num_columns;
	queue = GDAM_PRODUCER_QUEUE(GDAM_SKIN(skin)->model);
	if (skin->track_list){
		g_warning("Already have queue clist.");
		return;
	}
	skin->track_list = clist;
	//gtk_clist_set_reorderable(clist, TRUE);
	gtk_signal_connect_object(GTK_OBJECT(clist),
			   "row_move",
			   GTK_SIGNAL_FUNC(on_clist_row_move),
			   GTK_OBJECT(skin));
	gtk_signal_connect(GTK_OBJECT(queue),
			   "set_queue_list",
			   GTK_SIGNAL_FUNC(on_queue_list_set),
			   skin);
	gdam_widget_string_dest_set(GTK_WIDGET(clist),
				    on_filename_received,
				    "gdam-filename-text",
				    skin);
	num_columns = skin->track_list->columns;
	skin->column_fillers  = g_new(QListEntryGetter, num_columns);
	for (i=0; i<num_columns; i++) {
		char* title = column_names[i];
		GtkJustification just;
		if (title == NULL)
			break;
		gdam_debug("filler for column %d based on %s", i, title);
		skin->column_fillers[i] = get_column_filler(title);
		just = get_column_justification(title);
		gtk_clist_set_column_justification(clist, i, just);
	}
	for (; i<num_columns; i++)
		skin->column_fillers[i] = get_column_filler(NULL);
}


static void do_import_files(GdamProducerQueueSkin* skin)
{
	GdamSongSelectorSkin* songselskin;
	GdamSkin* packer;
	GdamProducerQueue* queue;
	char* cur_file;
	char* filenames;
	char* nl;
	gboolean keep_going = TRUE;

	queue = GDAM_PRODUCER_QUEUE(GDAM_SKIN(skin)->model);

	packer = gdam_skin_try_get_unique("GdamSongSelector");
	if (!packer){
		gdam_debug("producermp3skin: no songsel open, can't import.");
		return;
	}
	songselskin = GDAM_SONG_SELECTOR_SKIN( 
			gdam_packer_get_skin(GDAM_PACKER(packer), "songsel"));
	filenames = gdam_song_selector_export_selection_string(songselskin);
	cur_file = filenames;
	while (keep_going && *cur_file != '\0'){
		nl = strchr(cur_file, '\n');
		if (nl)
			*nl = '\0';
		else
			keep_going = FALSE;
		gdam_debug("filename %s dropped", cur_file);
		gdam_producer_queue_add_filename(queue, cur_file);
		cur_file = nl+1;
	}
	refresh_track_list(skin, queue->songinfo_list);
	g_free(filenames);
	return;
}


static gboolean gdam_queue_delete_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	connect_button_with_mask(GDAM_PRODUCER_QUEUE_SKIN(skin),
				 object,
				 GTK_SIGNAL_FUNC(on_delete_clicked),
				 skin);
	return TRUE;
}

static gboolean gdam_queue_up_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	connect_button_with_mask(GDAM_PRODUCER_QUEUE_SKIN(skin),
				 object,
				 GTK_SIGNAL_FUNC(on_up_clicked),
				 skin);
	return TRUE;
}

static gboolean gdam_queue_down_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	connect_button_with_mask(GDAM_PRODUCER_QUEUE_SKIN(skin),
				 object,
				 GTK_SIGNAL_FUNC(on_down_clicked),
				 skin);
	return TRUE;
}

static gboolean gdam_queue_refresh_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	connect_button_with_mask(GDAM_PRODUCER_QUEUE_SKIN(skin),
				 object,
				 GTK_SIGNAL_FUNC(on_refresh_clicked),
				 skin);
	return TRUE;
}

static gboolean gdam_queue_select_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	connect_button_with_mask(GDAM_PRODUCER_QUEUE_SKIN(skin),
				 object,
				 GTK_SIGNAL_FUNC(on_select_clicked),
				 skin);
	return TRUE;
}

static gboolean gdam_queue_clist_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	if (!GTK_IS_CLIST(object)) {
		g_warning("%s only valid for CList's", handler_names[0]);
		return TRUE;
	}
	connect_clist(GDAM_PRODUCER_QUEUE_SKIN(skin),
	              handler_names + 1,
		      //NULL,
		      GTK_CLIST(object));
	return TRUE;
}

static gboolean gdam_queue_skip_song_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	int num;
	SkipInfo* skip_info;
	if (handler_names[1] == NULL) {
		g_warning("%s needs an integer argument", handler_names[0]);
		return TRUE;
	}
	num = (int)atof(handler_names[1]);
	skip_info = g_new(SkipInfo, 1);
	skip_info->queue = GDAM_PRODUCER_QUEUE(skin->model);
	skip_info->num = num;
	connect_button_with_mask(GDAM_PRODUCER_QUEUE_SKIN(skin),
				 object,
				 GTK_SIGNAL_FUNC(on_skip_clicked),
				 skip_info);
	return TRUE;
}

static gboolean gdam_queue_filename_entry_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(GTK_OBJECT(skin->model),
				  "set_filename",
				  GTK_SIGNAL_FUNC(gtk_entry_set_text),
				  object);
	return TRUE;
}

static int queue_delete_event(GtkWidget* widget, GdkEvent* event,
			       GdamProducerQueueSkin* skin)
{
	toggle_queue_visibility(skin);
	return TRUE;
}

static gboolean gdam_queue_import_songsel_files_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
			   	  signal_name,
			   	  GTK_SIGNAL_FUNC(do_import_files),
			   	  GTK_OBJECT(skin));
	return TRUE;
}

static gboolean gdam_queue_queue_toplevel_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamProducerQueueSkin* queue_skin = GDAM_PRODUCER_QUEUE_SKIN(skin);
	queue_skin->queue_toplevel = GTK_WIDGET(object);
	gtk_object_ref(object);
	gtk_signal_connect(object,
			   "delete_event",
			   GTK_SIGNAL_FUNC(queue_delete_event),
			   skin);
	gdam_debug("Queue toplevel set.");
	return TRUE;
}

static gboolean gdam_queue_repeat_toggle_button_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
			   	 signal_name,
			   	 GTK_SIGNAL_FUNC(repeat_toggled),
			   	 GTK_OBJECT(skin));
	return TRUE;
}

static gboolean gdam_queue_toggle_queue_visibility_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
			   	 signal_name,
			   	 GTK_SIGNAL_FUNC(toggle_queue_visibility),
			   	 GTK_OBJECT(skin));
	return TRUE;
}

static void on_set_filename(GdamProducerQueueSkin* skin){
	GdamProducerQueue* queue;
	queue = GDAM_PRODUCER_QUEUE(GDAM_SKIN(skin)->model);
	refresh_track_list(skin, queue->songinfo_list);
}

static gboolean       gdam_producer_queue_build_new_model
                                                 (GdamSkin*      skin,
                                                  GdamChannel*   channel,
						  int 		 num_options,
						  char**         options)
{
	GdamModel* model;
	model = GDAM_MODEL(gtk_type_new(GDAM_TYPE_PRODUCER_QUEUE));
	model->channel = channel;
	skin->model = model;
	gtk_signal_connect_object(GTK_OBJECT(model),
				  "set_filename",
		           	  on_set_filename,
			   	  GTK_OBJECT(skin));
	return TRUE;
}

static void on_queue_filename_dragged(GtkSelectionData* selection_data,
				      gpointer 		data)
{
	GdamProducerQueue* queue = GDAM_PRODUCER_QUEUE(data);
	const char* fname;
	if (!queue->current_song)
		return;
	fname = gdam_song_info_get_filename(queue->current_song), 
	gtk_selection_data_set (selection_data, selection_data->target, 8,
				(guchar*)strdup(fname), 
				strlen(fname));
}

static gboolean gdam_queue_frame_drag_data_get_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gdam_debug("producerqueue: setting filename source");
	gdam_widget_string_source_set(
				  GTK_WIDGET(object),
				  on_queue_filename_dragged,
				  "gdam-filename-text",
				  GDAM_PRODUCER_QUEUE(skin->model));
	return TRUE;
}


static GdamSkinGenericConnect connect_infos[] =
{
	{ "gdam_queue_filename_drag_received",
	   gdam_queue_filename_drag_received_handler	},
	{ "gdam_queue_clist",
	   gdam_queue_clist_handler			},
	{ "gdam_queue_delete_clicked",
	   gdam_queue_delete_handler			},
	{ "gdam_queue_up_clicked",
	   gdam_queue_up_handler			},
	{ "gdam_queue_down_clicked",
	   gdam_queue_down_handler			},
	{ "gdam_queue_refresh_clicked",
	   gdam_queue_refresh_handler			},
	{ "gdam_queue_select_clicked",
	   gdam_queue_select_handler			},
	{ "gdam_queue_filename_entry",
	   gdam_queue_filename_entry_handler		},
	{ "gdam_queue_drag_data_get",
	   gdam_queue_frame_drag_data_get_handler	},
	{ "gdam_queue_queue_toplevel",
	   gdam_queue_queue_toplevel_handler		},
	{ "gdam_queue_import_songsel_files",
	   gdam_queue_import_songsel_files_handler	},
	{ "gdam_queue_toggle_queue_visibility",
	   gdam_queue_toggle_queue_visibility_handler	},
	{ "gdam_queue_repeat_toggle_button",
	   gdam_queue_repeat_toggle_button_handler	},
	{ "gdam_queue_skip_song",
	   gdam_queue_skip_song_handler			}
};

static void gdam_producer_queue_destroy(GtkObject* object){
	GdamProducerQueueSkin* skin = GDAM_PRODUCER_QUEUE_SKIN(object);
	if (skin->queue_toplevel){
		GtkObject* object = GTK_OBJECT(skin->queue_toplevel);
		skin->queue_toplevel = NULL;
		gtk_object_unref(object);
	}
	parent_class->destroy(object);
}

static void gdam_producer_queue_skin_class_init(GdamSkinClass* skin_class)
{
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(skin_class);
	skin_class->build_new_model
				= gdam_producer_queue_build_new_model;
	object_class->destroy = gdam_producer_queue_destroy;
	gdam_skin_class_register_generic_handlers(skin_class,
				                  G_N_ELEMENTS(connect_infos),
				                  connect_infos);
}

